local cm = {} -- Mtatable pour classe complex
function complex(a, b) -- Constructeur
  b = b or 0
  assert(type(a) == "number" and type(b) == "number")
  return setmetatable({a=a, b=b}, cm)
end
function cm.__tostring(z)
  return ("%.14g%+.14gi"):format(z.a,z.b)
end
function cm.__add(z1, z2)
  return complex(z1.a+z2.a, z1.b+z2.b)
end
function cm.__sub(z1, z2)
  return complex(z1.a-z2.a, z1.b-z2.b)
end
--------------------
function cm.__mul(z1, z2)
  return complex(z1.a*z2.a-z1.b*z2.b, z1.a*z2.b+z1.b*z2.a)
end
function cm.__div(z1, z2)
  local den = z2.a^2+z2.b^2
  return complex((z1.a*z2.a+z1.b*z2.b)/den,
    (z1.b*z2.a-z1.a*z2.b)/den)
end
function cm.__unm(z)
  return complex(-z.a,-z.b)
end
-- L'opration modulo n'est pas dfinie pour les complexes !
function cm.__mod(z1, z2)
  return complex(z1.a%z2.a, z1.b%(z2.b ~= 0 and z2.b or 1))
end
--------------------------
function cm.__index(z, k)
  if k=='r' then
    return math.sqrt(z.a^2+z.b^2)
  elseif k== 'th' then
    return math.atan2(z.b, z.a)
  else
    return cm[k]
  end
end
function cm.__newindex(z,k,n)
  if k=='r' then
    local mult=n/z.r
    z.a=z.a*mult
    z.b=z.b*mult
  elseif k== 'th' then
    local r = z.r
    z.a=r*math.cos(n)
    z.b=r*math.sin(n)
  else
    rawset(z, k, n)
  end
end
function cm.__pow(z1, z2)
  local r, th = z1.r, z1.th
  local z = complex(r^z2.a*math.exp(-z2.b*th))
  z.th = z2.b*math.log(r) + z2.a*th
  return z
end
-----------------------------
function cm.__len(z)
  return math.sqrt(z.a^2+z.b^2)
end
function cm.__concat(z1, z2)
  return tostring(z1)..tostring(z2)
end
function cm.__gc(z)
--  print('INFO: '..z..' collected')
end
cm.__metatable = 'Complex number'
--------------------------------
debug.setmetatable(0, {__index=function(v,k)
  return ({a=v, r=math.abs(v), b=0, th=0})[k] or math[k]
end})
----------------
function cm.__eq(z1, z2)
  return z1.a==z2.a and z1.b==z2.b
end
function cm.__lt(z1, z2)
  return z1.r < z2.r
end
function cm.__le(z1, z2)
  return z1.r <= z2.r
end
-----------------------------------
function cm.__call(z, coord)
  if coord == 'cartesian' then
    return z.a, z.b
  elseif coord == 'polar' then
    return z.r, z.th
  elseif coord == nil then
    return z.a, z.b, z.r, z.th
  end
end
function cm.__pairs(z)
  return pairs{a=z.a, b=z.b, r=z.r, th=z.th}
end
function cm.__ipairs(z)
  local i = complex(0,1)
  return function(zend, z)
    local t={z-1, z+1, z-i, z+i}
    local mindiff = #(zend - z)
    local res
    for _,v in ipairs(t) do
      local diff = #(zend - v)
      if diff < mindiff and #v <= #zend then
         mindiff = diff
         res = v
      end
    end
    if res then
      return res, res.a, res.b
    end
  end, z, complex(0)
end
------
function cm.exp(z)
  local z2=complex(math.exp(z.a))
  z2.th = z.b
  return z2
end
local i = complex(0, 1)
function cm.sin(z)
  return (cm.exp(i*z)-cm.exp(-i*z))/(2*i)
end
function cm.cos(z)
  return (cm.exp(i*z)+cm.exp(-i*z))/2
end
function cm.tan(z)
  return z:sin()/z:cos()
end
return complex
